/***************************************************************************
                          htl_contour.cpp  -  description
                             -------------------
    begin                : Sat Oct 4 2003
    copyright            : (C) 2003 by Soloviev Alexei,JINR,LIT-303,64713
    email                : solovjev@cv.jinr.ru
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "htl_contour.h"
#include "gl2ps.h"

#include <algorithm>
#include <cmath>
using namespace std;

#include <GL/gl.h>

const double & htl::contour::xmin () const
{
  if ( logX() ) {
    const_key_iterator p = 
      find_if ( dbegin(), dend(), bind2nd ( greater<double>(), 0 ) );
    return p != dend() ? *p : right();
  }
  return left();
}
const double & htl::contour::xmax () const
{
  return right();
}

const double & htl::contour::ymin () const
{
  if ( logY() ) {
    static double ymin4log;
    ymin4log = ymax();
    for ( unsigned int j = 0; j < size(); ++j ) {
      if ( at(j) > 0 && at(j) < ymin4log ) ymin4log = at(j);
    }
    return ymin4log;
  }
  return *min_element ( begin(), end() );
}

const double & htl::contour::ymax () const
{
  return *max_element ( begin(), end() );
}

void htl::contour::plot () const
{
  // color
  glColor3ub ( _color.r, _color.g, _color.b );

  // blending
  glEnable ( GL_BLEND );
  glBlendFunc ( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );

  // point
  glEnable ( GL_POINT_SMOOTH );
  glHint ( GL_POINT_SMOOTH_HINT, GL_NICEST );
  glPointSize ( 4 );
  gl2psPointSize ( 4 );

  // line
  glEnable ( GL_LINE_SMOOTH );
  glHint ( GL_LINE_SMOOTH_HINT, GL_DONT_CARE );
  glLineWidth ( 1.5 );
  gl2psLineWidth ( 1.5 );

  switch ( _style ) {
  case lines:
    plot_lines();
    break;
  case points:
    plot_points();
    break;
  case linespoints:
    plot_lines();
    plot_points();
    break;
  case dots:
    glPointSize ( 2 );
    gl2psPointSize ( 2 );
    plot_points();
    break;
  case impulses:
    plot_impulses();
    break;
  case steps:
    plot_steps();
    break;
  case boxes:
    plot_boxes();
    break;
  case candlesticks:
    glLineWidth ( 3 );
    gl2psLineWidth ( 3 );
    plot_impulses();
    break;
  case errorbars:
    plot_points();
    plot_errorbars();
    break;
  default:
    plot_lines();
    break;
  }
}

void htl::contour::plot_lines () const
{
  for ( unsigned int j = 1; j < size(); ++j ) {
    double x1 = bincenter(j-1);
    double x2 = bincenter(j);
    double y1 = at(j-1);
    double y2 = at(j);
    if ( logX() ) {
      if ( x1 < xmin() || x2 < xmin() ) continue;
      x1 = log10 ( x1 );
      x2 = log10 ( x2 );
    }
    if ( logY() ) {
      if ( y1 < ymin() || y2 < ymin() ) continue;
      y1 = log10 ( y1 );
      y2 = log10 ( y2 );
    }
    glBegin ( GL_LINES );
      glVertex2f ( x1, y1 );
      glVertex2f ( x2, y2 );
    glEnd ();
  }
}

void htl::contour::plot_points () const
{
  for ( unsigned int j = 0; j < size(); ++j ) {
    double x = bincenter(j);
    double y = at(j);
    if ( logX() ) {
      if ( x < xmin() ) continue;
      x = log10 ( x );
    }
    if ( logY() ) {
      if ( y < ymin() ) continue;
      y = log10 ( y );
    }
    glBegin ( GL_POINTS );
      glVertex2f ( x, y );
    glEnd ();
  }
}

void htl::contour::plot_impulses () const
{
  for ( unsigned int j = 0; j < size(); ++j ) {
    double x = bincenter(j);
    double y = at(j);
    double y0 = logY() ? ymin() : 0;
    if ( logX() ) {
      if ( x < xmin() ) continue;
      x = log10 ( x );
    }
    if ( logY() ) {
      if ( y < ymin() ) continue;
      y = log10 ( y );
      y0 = log10 ( y0 );
    }
    glBegin ( GL_LINES );
      glVertex2f ( x, y );
      glVertex2f ( x, y0 );
    glEnd ();
  }
}

void htl::contour::plot_steps () const
{
  for ( unsigned int j = 0; j < size(); ++j ) {
    double x1 = bincenter(j) - binwidth(j) / 2;
    double x2 = bincenter(j) + binwidth(j) / 2;
    double y = at(j);
    double y1 = at(j>0?j-1:j);
    if ( logX() ) {
      if ( x1 < xmin() || x2 < xmin() ) continue;
      x1 = log10 ( x1 );
      x2 = log10 ( x2 );
    }
    if ( logY() ) {
      if ( y < ymin() || y1 < ymin() ) continue;
      y = log10 ( y );
      y1 = log10 ( y1 );
    }
    glBegin ( GL_LINE_STRIP );
      glVertex2f ( x1, y1 );
      glVertex2f ( x1, y );
      glVertex2f ( x2, y );
    glEnd ();
  }
}

void htl::contour::plot_boxes () const
{
  for ( unsigned int j = 0; j < size(); ++j ) {
    double x1 = bincenter(j) - binwidth(j) / 2;
    double x2 = bincenter(j) + binwidth(j) / 2;
    double y = at(j);
    double y0 = logY() ? ymin() : 0;
    if ( logX() ) {
      if ( x1 < xmin() || x2 < xmin() ) continue;
      x1 = log10 ( x1 );
      x2 = log10 ( x2 );
    }
    if ( logY() ) {
      if ( y < ymin() ) continue;
      y = log10 ( y );
      y0 = log10 ( y0 );
    }
    glBegin ( GL_LINE_STRIP );
      glVertex2f ( x1, y0 );
      glVertex2f ( x1, y );
      glVertex2f ( x2, y );
      glVertex2f ( x2, y0 );
    glEnd ();
  }
}

void htl::contour::plot_errorbars () const
{
  for ( unsigned int j = 0; j < size(); ++j ) {
    double x = bincenter(j);
    double y = at(j);
    double y1 = y - at(j).error();
    double y2 = y + at(j).error();
    if ( logX() ) {
      if ( x < xmin() ) continue;
      x = log10 ( x );
    }
    if ( logY() ) {
      if ( y < ymin() ) continue;
      y = log10 ( y );
      y1 = log10 ( max ( y1, ymin() ) );
      y2 = log10 ( y2 );
    }
    glBegin ( GL_LINES );
      glVertex2f ( x, y1 );
      glVertex2f ( x, y2 );
    glEnd ();
  }
}

const htl::contour::rgb htl::contour::_web_color_rgbs[96] = 
{
  { 0xff, 0x00, 0x00 },               /*   color 01: red           */
  { 0x00, 0xc0, 0x00 },               /*   color 02: dark green    */
  { 0x00, 0x80, 0xff },               /*   color 03: dark blue     */
  { 0xc0, 0x00, 0xff },               /*   color 04: dark magenta  */
  { 0xc0, 0xff, 0x40 },               /*   color 05: yellow        */
  { 0xc0, 0x40, 0x00 },               /*   color 06: orange        */
  { 0x40, 0xff, 0x80 },               /*   color 07: sea green     */
  { 0x20, 0x20, 0xc0 },               /*   color 08: royal blue    */
  { 0x80, 0x00, 0xc0 },               /*   color 09: dark violet   */
  { 0x00, 0x60, 0x80 },               /* DeepSkyBlue4 */
  { 0x00, 0x80, 0x00 },               /* green4 */
  { 0x00, 0x80, 0x40 },               /* SpringGreen4 */
  { 0x00, 0x80, 0x80 },               /* dark cyan, turquoise4 */
  { 0x00, 0xc0, 0x60 },               /* SpringGreen3 */
  { 0x00, 0xc0, 0xc0 },               /* cyan3, turquoise3 */
  { 0x00, 0xff, 0x00 },               /* green */
  { 0x20, 0x80, 0x20 },               /* forest green */
  { 0x30, 0x60, 0x80 },               /* SteelBlue4 */
  { 0x40, 0x40, 0x40 },               /* grey25-31 */
  { 0x40, 0x80, 0x00 },               /* chartreuse4 */
  { 0x00, 0x00, 0x80 },               /* dark blue, navy blue */
  { 0x80, 0x60, 0x00 },               /* DarkGoldenrod4 */
  { 0x80, 0x60, 0x10 },               /* goldenrod4 */
  { 0x80, 0x60, 0x60 },               /* pink4 */
  { 0x80, 0x60, 0x80 },               /* plum4 */
  { 0x00, 0x00, 0xc0 },               /* medium blue */
  { 0x00, 0x00, 0xff },               /* blue */
  { 0x00, 0x60, 0x00 },               /* dark green */
  { 0x40, 0xc0, 0x80 },               /* SeaGreen3 */
  { 0x60, 0xa0, 0xc0 },               /* SkyBlue3 */
  { 0x60, 0xc0, 0x00 },               /* chartreuse3 */
  { 0x60, 0xc0, 0xa0 },               /* medium aquamarine */
  { 0x80, 0x00, 0x00 },               /* dark red */
  { 0x80, 0x00, 0x80 },               /* dark magenta */
  { 0x60, 0x20, 0x80 },               /* DarkOrchid4 */
  { 0x60, 0x60, 0x60 },               /* dim grey */
  { 0x00, 0xff, 0xff },               /* cyan1, turquoise1 */
  { 0x20, 0x20, 0x20 },               /* grey13-18 */
  { 0x20, 0x40, 0x40 },               /* dark slate grey */
  { 0x20, 0x40, 0x80 },               /* RoyalBlue4 */
  { 0x60, 0x80, 0x20 },               /* olive drab */
  { 0x60, 0x80, 0x60 },               /* DarkSeaGreen4 */
  { 0x60, 0x80, 0x80 },               /* LightBlue4, PaleTurquoise4 */
  { 0x80, 0x80, 0x40 },               /* LightGoldenrod4, khaki4 */
  { 0x80, 0x80, 0x80 },               /* grey51-56 */
  { 0xa0, 0xa0, 0xa0 },               /* dark grey, grey63-68 */
  { 0xa0, 0xd0, 0xe0 },               /* light blue */
  { 0xc0, 0x20, 0x20 },               /* firebrick3 */
  { 0xc0, 0x60, 0x00 },               /* DarkOrange3 */
  { 0x80, 0xc0, 0xe0 },               /* sky blue */
  { 0xc0, 0x60, 0xc0 },               /* orchid3 */
  { 0xc0, 0x80, 0x00 },               /* orange3 */
  { 0xc0, 0x80, 0x60 },               /* LightSalmon3 */
  { 0xff, 0x40, 0x00 },               /* orange red */
  { 0xff, 0x40, 0x40 },               /* brown1, tomato */
  { 0x80, 0xc0, 0xff },               /* light sky blue */
  { 0xff, 0x80, 0x60 },               /* salmon */
  { 0xff, 0x80, 0x80 },               /* light coral */
  { 0xc0, 0xa0, 0x00 },               /* gold3 */
  { 0xc0, 0xc0, 0xc0 },               /* grey76-81, honeydew3, ivory3, snow3 */
  { 0xc0, 0xff, 0xc0 },               /* DarkSeaGreen1 */
  { 0xff, 0x00, 0x00 },               /* red */
  { 0xff, 0x00, 0xff },               /* magenta */
  { 0xff, 0x80, 0xa0 },               /* PaleVioletRed1 */
  { 0xff, 0x80, 0xff },               /* orchid1 */
  { 0xc0, 0xc0, 0xa0 },               /* LemonChiffon3 */
  { 0xff, 0x60, 0x60 },               /* IndianRed1 */
  { 0xff, 0x80, 0x00 },               /* dark orange */
  { 0xff, 0xa0, 0x00 },               /* orange */
  { 0x80, 0xe0, 0xe0 },               /* CadetBlue2, DarkSlateGray2 */
  { 0xa0, 0xe0, 0xe0 },               /* pale turquoise */
  { 0xa0, 0xff, 0x20 },               /* green yellow */
  { 0xc0, 0x00, 0x00 },               /* red3 */
  { 0xc0, 0x00, 0xc0 },               /* magenta3 */
  { 0xa0, 0x20, 0x20 },               /* brown */
  { 0xa0, 0x20, 0xff },               /* purple */
  { 0x80, 0x20, 0x00 },               /* OrangeRed4 */
  { 0x80, 0x20, 0x20 },               /* brown4 */
  { 0x80, 0x40, 0x00 },               /* DarkOrange4 */
  { 0x80, 0x40, 0x20 },               /* sienna4 */
  { 0x80, 0x40, 0x80 },               /* orchid4 */
  { 0x80, 0x60, 0xc0 },               /* MediumPurple3 */
  { 0x80, 0x60, 0xff },               /* SlateBlue1 */
  { 0x80, 0x80, 0x00 },               /* yellow4 */
  { 0xa0, 0x80, 0xff },               /* MediumPurple1 */
  { 0xc0, 0x60, 0x80 },               /* PaleVioletRed3 */
  { 0xc0, 0xc0, 0x00 },               /* yellow3 */
  { 0xff, 0x80, 0x40 },               /* sienna1 */
  { 0xff, 0xa0, 0x40 },               /* tan1 */
  { 0xff, 0xa0, 0x60 },               /* sandy brown */
  { 0xff, 0xa0, 0x70 },               /* light salmon */
  { 0xff, 0xc0, 0x20 },               /* goldenrod1 */
  { 0xff, 0xc0, 0xc0 },               /* RosyBrown1, pink */
  { 0xff, 0xff, 0x00 },               /* yellow */
  { 0xff, 0xff, 0x80 },               /* khaki1 */
  { 0xff, 0xff, 0xc0 }                /* lemon chiffon */
};

unsigned int htl::contour::_color_index;

